#include <math.h>
#include <stdlib.h>
#include <iostream.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>


void HSVToRGB(double h, double s, double v, double &r,
              double &g, double &b){
  // adapted from http://www.research.compaq.com/SRC/m3sources/html
  //                    /color/src/Color.m3.html#FromHSV

  if (v==0.0){
    r=g=b=0.0;
    return; // Black
  }
  if (s==0.0){
    r=g=b=v;
    return; // Gray
  }
  if (h==1.0) h=0.0;
  int i=(int)(h*6.0);
  double f=h*6.0-i;
  double p=v*(1.0-s);
  double q=v*(1.0-s*f);
  double t=v*(1.0-s*(1.0-f));
  switch(i){
  case 0: r=v; g=t; b=p;
    break;
  case 1: r=q; g=v; b=p;
    break;
  case 2: r=p; g=v; b=t;
    break;
  case 3: r=p; g=q; b=v;
    break;
  case 4: r=t; g=p; b=v;
    break;
  case 5: r=v; g=p; b=q;
    break;
  }
  
  return;  
}

double r, g, b;

SoSeparator *
makeScene ()
  {

  SoSeparator *scene = new SoSeparator();

  SoSeparator *mySep;
  SoTransform *myXform;
  SoSphere *mySphere;
  SoMaterial *myMat;
  int i, j, k;
  SbVec3f x, y;

  for (i=1; i<=4; i++)
    for (j=1; j<=4; j++)
      for (k=1; k<=4; k++){
	mySep = new SoSeparator();
        myXform = new SoTransform();
        mySphere = new SoSphere();

	x = SbVec3f(i-2.5, j-2.5, k-2.5);
	y = x;
	y.normalize();
	x = x + y*10.0;
	myXform->translation.setValue(x);
	mySphere->radius.setValue(x.length()/5.0);
	myMat = new SoMaterial();

	HSVToRGB(i/4.0,
             j/4.0,
             k/8.0,
             r,g,b);

	myMat->emissiveColor.setValue(r, g, b);

	mySep->addChild(myXform);
	mySep->addChild(myMat);
	mySep->addChild(mySphere);
	scene->addChild(mySep);
      }
  
  
  return scene;
  }

float PresentTime = 0.0;
float DTime = 0.015;

void
animate(void *, SoSensor *)
{
  PresentTime += 3.14159/20.0; //DTime;
}

////////////////////////////////////////////////////////////
//
// The main program gets a window from X, 
// puts an interactive 3D examinerViewer in it, 
// attaches the scene graph to the viewer,
// renders the scene graph,
// and hands control over to the Inventor main loop.

void
main(int, char **argv)
  {

  // Initialize Inventor and Xt
  Widget myWindow = SoXt::init(argv[0]);
  if (myWindow == NULL) exit(1);

  // SoGLRenderAction *rend = new SoGLRenderAction();
  //rend->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_ADD);
  
  // Create a scene graph
  SoSeparator *root = makeScene();

  SoTimerSensor *sensor = new SoTimerSensor(animate, root);
  sensor->setInterval(SbTime(0.015)); // 70 frames per second
  sensor->schedule();
  
  SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->setTitle("Simple Scene");
  myViewer->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_BLEND);
  myViewer->show();

  SoWriteAction wa;  // Write out the scene graph just for demo purposes 
  wa.apply(root);    // The scenegraph can be displayed using

  SoXt::show(myWindow);
  SoXt::mainLoop();
  }
