// Kevin Beason
// http://www.cs.fsu.edu/~beason/
// 4/16/02
// 
// Update Emittance and Display Scene
// Sphere's surface points are moved around, and are colored
// to match a cosine emittance distribution.
// 
// Also, a scene read in from a file is superimposed.


#include <math.h>
#include <stdlib.h>
#include <iostream.h>
#include <string.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>

void HSVToRGB(double h, double s, double v, double &r,
	      double &g, double &b){
  // adapted from http://www.research.compaq.com/SRC/m3sources/html
  //                    /color/src/Color.m3.html#FromHSV

  if (v==0.0){
    r=g=b=0.0;
    return; // Black
  }
  if (s==0.0){
    r=g=b=v;
    return; // Gray
  }
  if (h==1.0) h=0.0;
  int i=(int)(h*6.0);
  double f=h*6.0-i;
  double p=v*(1.0-s);
  double q=v*(1.0-s*f);
  double t=v*(1.0-s*(1.0-f));
  switch(i){
  case 0: r=v; g=t; b=p;
    break;
  case 1: r=q; g=v; b=p;
    break;
  case 2: r=p; g=v; b=t;
    break;
  case 3: r=p; g=q; b=v;
    break;
  case 4: r=t; g=p; b=v;
    break;
  case 5: r=v; g=p; b=q;
    break;
  }
  
  return;  
}

SoPointLight *light;
const int dim = 4000;
int numSamples = 400;
double myGamma = 0.5;
SoMaterial *myMats[dim];
SoTransform *myXforms[dim];
SoCamera *Camera;
double r, g, b;


SoSeparator *
makeScene ()
  {

  SoSeparator *scene = new SoSeparator();

  double t;
  int i, j, k;
  double x, y, z;
  //double l, w, h;
  double xMin=-1.0, xMax=1.0;
  SoTransform *xform;
  SoCube *cube;
  SoMaterial *mat;
  SoSeparator *sep;
  SoSphere *sphere;
  double rad, angle;
  SbVec3f point;

  SoComplexity *complexity = new SoComplexity();
  complexity = new SoComplexity();
  complexity->value = 0.30;
  scene->addChild(complexity);

  // read in scene
  SoInput *myInput = new SoInput();
  SoSeparator *myGraph;
  if (myInput->openFile("scene.iv")){
    myGraph = SoDB::readAll(myInput);
    myInput->closeFile();
    scene->addChild(myGraph);
  }
  else {
    cerr << "Error opening scene graph file.\n";
  }
      
    HSVToRGB(0.5,
	     0.5,
	     0.5,
	     r,g,b);

     
  for (i=0; i<numSamples; i++){ // loop through angles
    //l = 100.0/numSamples; //drand48() *0.5;
    point = SbVec3f(drand48()*2.0-1.0,
		    drand48()*2.0-1.0,
		    drand48()*2.0-1.0);
    while (point.length()>1.0){
        point = SbVec3f(drand48()*2.0-1.0,
			drand48()*2.0-1.0,
			drand48()*2.0-1.0);
    }
    point.normalize();
    x=point[0];
    y=point[1];
    z=point[2];
     
    // separator node
    sep = new SoSeparator();
    
    // transparency node
    mat = new SoMaterial();
    mat->ambientColor.setValue(0.0, 0.0, 0.0);
    mat->emissiveColor.setValue(r, g, b);
    mat->diffuseColor.setValue(0.0, 0.0, 0.0);
    mat->specularColor.setValue(0.0, 0.0, 0.0);
    mat->shininess.setValue(0.0);
    mat->transparency.setValue(0.0);
    myMats[i] = mat;

    // transform node
    xform = new SoTransform();
    xform->translation.setValue(x, y, z);
    myXforms[i] = xform;
    
    // sphere node
    sphere = new SoSphere();
    sphere->radius.setValue(4.0/sqrt(numSamples));

    // make it
    sep->addChild(mat);
    sep->addChild(xform);
    sep->addChild(sphere);
    scene->addChild(sep);

  }
  

  return scene;
  }

float PresentTime = 0.0;
float DTime = 0.015;

int moveIt=0;

void
animate(void *, SoSensor *)
{
  int i;
  double v;
  SbVec3f a, b, point;

  // move a sample
  moveIt++;
  if (moveIt>=numSamples) moveIt=0;
  point.setValue(1.0, 1.0, 1.0);
  while (point.length()>1.0){
    point = SbVec3f(drand48()*2.0-1.0,
		    drand48()*2.0-1.0,
		    drand48()*2.0-1.0);
  }
  point.normalize();
  myXforms[i]->translation.setValue(point[0], point[1], point[2]);

  // set emittance
  for (i=0; i<numSamples; i++){
     a=myXforms[i]->translation.getValue();
     a.normalize();
     b=Camera->position.getValue();
     b.normalize();
     v=a.dot(b);
     myMats[i]->emissiveColor.setValue(v, v, v);
  }
  PresentTime += 3.14159/20.0; //DTime;
}

// This code is from Dr. David C. Banks
// http://www.cs.fsu.edu/~banks/

////////////////////////////////////////////////////////////
//
// The main program gets a window from X, 
// puts an interactive 3D examinerViewer in it, 
// attaches the scene graph to the viewer,
// renders the scene graph,
// and hands control over to the Inventor main loop.

void
main(int argc, char* argv[])
{

  // read command line arguments
  int i, j;
  for (i=1; i<argc; i++){
    if (strcmp(argv[i], "-s") == 0)
      numSamples = atoi(argv[i+1]);
    else if (strcmp(argv[i], "-g") == 0)
      myGamma = atof(argv[i+1]);
    else if (strcmp(argv[i], "-h") == 0){
      cerr << "\nUsage: " << argv[0]
	   << " [ -s samples ] [ -g gamma ]\n\n";
      exit(1);
    }
  }

    
  // Initialize Inventor and Xt
  Widget myWindow = SoXt::init(argv[0]);
  if (myWindow == NULL) exit(1);

  // SoGLRenderAction *rend = new SoGLRenderAction();
  //rend->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_ADD);
  
  // Create a scene graph
  SoSeparator *root = makeScene();

  SoTimerSensor *sensor = new SoTimerSensor(animate, root);
  sensor->setInterval(SbTime(0.015)); // 70 frames per second
  sensor->schedule();
  
  SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  Camera = myViewer->getCamera();
  myViewer->setTitle("Update Emittance and Display Scene");
  myViewer->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_BLEND);
  myViewer->show();

  SoWriteAction wa;  // Write out the scene graph just for demo purposes 
  //wa.apply(root);    // The scenegraph can be displayed using
		     // ivview, SceneViewer, or gview .
		     // It can also be displayed in a Web browser by
		     // changing its header from
		     //
		     // #Inventor V2.1 ascii
		     // 
		     // to
		     //
		     // #VRML V1.0 ascii
		     //
		     // to exploit the (deprecated) original version
		     // of the Virtual Reality Modelling Language
		     // file format.
		     // 
		     // Visit http://www.vrml.org/ (deprecated URL) or
		     // http://www.web3d.org/ for info on 
		     // VRML begat Web3D begat X3D.


  SoXt::show(myWindow);
  SoXt::mainLoop();
  }
